import asyncio
import os
import random
import shutil
import string
import unittest

import aiobotocore.session

import pydatatask


def rid(n=6):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(n))


class TestMinio(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.docker_name = None
        self.docker_path = shutil.which("docker")
        self.minio_endpoint = os.getenv("PYDATATASK_TEST_MINIO_ENDPOINT")
        self.minio_username = os.getenv("PYDATATASK_TEST_MINIO_USERNAME", "minioadmin")
        self.minio_password = os.getenv("PYDATATASK_TEST_MINIO_PASSWORD", "minioadmin")
        self.minio_secure = os.getenv("PYDATATASK_MINIO_SECURE", "0").lower() not in (
            "0",
            "",
            "false",
        )
        self.test_id = rid()
        self.client = None
        self.bucket = "test-pydatatask-" + self.test_id

    async def asyncSetUp(self):
        if self.minio_endpoint is None:
            if self.docker_path is None:
                raise unittest.SkipTest("No minio endpoint configured and docker is not installed")
            port = random.randrange(0x4000, 0x8000)
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "run",
                "--rm",
                "--name",
                self.bucket,
                "-d",
                "-p",
                f"{port}:9000",
                "minio/minio:latest",
                "server",
                "/data",
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()
            if await p.wait() != 0:
                raise unittest.SkipTest("No minio endpoint configured and docker failed to launch minio/minio:latest")
            self.minio_endpoint = f"localhost:{port}"
            self.docker_name = self.bucket
            await asyncio.sleep(1)
        minio_session = aiobotocore.session.get_session()
        self.client = await minio_session.create_client(
            "s3",
            endpoint_url="http://" + self.minio_endpoint,
            aws_access_key_id=self.minio_username,
            aws_secret_access_key=self.minio_password,
        ).__aenter__()

    async def test_minio(self):
        repo = pydatatask.S3BucketRepository(lambda: self.client, self.bucket, prefix="weh/", extension=".weh")
        repo_yaml = pydatatask.YamlMetadataS3Repository(lambda: self.client, self.bucket, prefix="weh/")
        assert repr(repo)
        await repo.validate()
        await repo_yaml.validate()

        async with await repo.open("foo", "w") as fp:
            await fp.write("hello world")
        async with (await self.client.get_object(Bucket=self.bucket, Key="weh/foo.weh"))["Body"] as fp:
            assert await fp.read() == b"hello world"
        async with await repo.open("foo", "wb") as fp:
            await fp.write(b"hello world")
        async with await repo.open("foo", "rb") as fp:
            assert await fp.read() == b"hello world"
        async with await repo.open("foo", "r") as fp:
            assert await fp.read() == "hello world"
        assert [ident async for ident in repo] == ["foo"]
        assert await repo.contains("foo")
        assert not await repo.contains("bar")

        info = await repo.info("bar")
        assert str(info) == f"s3://{self.bucket}/weh/bar.weh"
        assert info.endpoint == "http://" + self.minio_endpoint

        await repo.delete("foo")
        assert [ident async for ident in repo] == []
        await repo.delete("bar")

        await repo_yaml.dump("foo", {"weh": 1})
        assert await repo_yaml.info("foo") == {"weh": 1}
        assert await repo_yaml.info("bar") == {}

    async def asyncTearDown(self):
        if self.client is not None:
            objects = [
                {"Key": obj["Key"]} for obj in (await self.client.list_objects(Bucket=self.bucket)).get("Contents", [])
            ]
            if objects:
                await self.client.delete_objects(
                    Bucket=self.bucket,
                    Delete={
                        "Objects": objects,
                    },
                )
            await self.client.delete_bucket(Bucket=self.bucket)
            await self.client.close()

        if self.docker_name is not None:
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "kill",
                self.docker_name,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()


if __name__ == "__main__":
    unittest.main()



import unittest

import pydatatask


class TestRelated(unittest.IsolatedAsyncioTestCase):
    async def test_related(self):
        translator = pydatatask.InProcessMetadataRepository({"456": "123"})

        base1 = pydatatask.InProcessBlobRepository({"123": b"asdf"})
        related1 = pydatatask.RelatedItemRepository(base1, translator)
        async with await related1.open("456", "rb") as fp:
            assert await fp.read() == b"asdf"
        await related1.delete("456")
        assert [x async for x in base1] == ["123"]
        assert [x async for x in related1] == ["456"]
        assert await related1.contains("456")
        assert not await related1.contains("lksfdlkfd")
        with self.assertRaises(LookupError):
            await related1.info("lkjfdslkj")
        with self.assertRaises(LookupError):
            await related1.open("lkjfdslkj", "w")

        base2 = pydatatask.InProcessMetadataRepository({"123": "weh!"})
        related2 = pydatatask.RelatedItemRepository(base2, translator, allow_deletes=True, prefetch_lookup=False)
        assert await related2.info("456") == "weh!"

        await related2.delete("456")
        assert [x async for x in base2] == []
        await related2.delete("lsdlkjfj")


if __name__ == "__main__":
    unittest.main()



import asyncio
import base64
import os
import random
import shutil
import string
import unittest
import warnings

import kubernetes_asyncio

import pydatatask


def rid(n=6):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(n))


class TestKube(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.minikube_profile = None
        self.minikube_path = shutil.which("minikube")
        self.kube_context = os.getenv("PYDATATASK_TEST_KUBE_CONTEXT")
        self.kube_namespace = os.getenv("PYDATATASK_TEST_KUBE_NAMESPACE", "default")
        self.test_id = rid()

    async def asyncSetUp(self):
        if self.kube_context is None:
            if self.minikube_path is None:
                raise unittest.SkipTest("No kube context specified and minikube is not installed")
            self.minikube_profile = f"pydatatask-test-{self.test_id}"
            p = await asyncio.create_subprocess_exec(
                self.minikube_path,
                "start",
                "--profile",
                self.minikube_profile,
                "--interactive=false",
                "--keep-context",
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            stdout, _ = await p.communicate()
            code = await p.wait()
            if code != 0:
                raise unittest.SkipTest(
                    f"No kube context specified and minikube failed to start. Logs below:\n\n{stdout.decode()}"
                )
            self.kube_context = self.minikube_profile

    async def test_kube(self):
        session = pydatatask.Session()

        await kubernetes_asyncio.config.load_kube_config(context=self.kube_context)

        cluster_quota = pydatatask.ResourceManager(pydatatask.Resources.parse("1", "1Gi"))

        @session.resource
        async def podman():
            podman = pydatatask.PodManager(
                f"test-{self.test_id}",
                self.kube_namespace,
            )
            yield podman
            await podman.close()

        repo0 = pydatatask.InProcessMetadataRepository({str(i): f"weh-{i}" for i in range(50)})
        repoDone = pydatatask.InProcessMetadataRepository()
        repoLogs = pydatatask.InProcessBlobRepository()

        if self.minikube_profile is not None:
            # wait for boot I guess
            async with kubernetes_asyncio.client.ApiClient() as api:
                v1 = kubernetes_asyncio.client.CoreV1Api(api)
                for i in range(200):
                    if (await v1.list_namespaced_service_account(self.kube_namespace)).items:
                        break
                    await asyncio.sleep(0.1)
                else:
                    raise Exception("Minikube failed to give us anything interesting within 20 seconds")

        task = pydatatask.KubeTask(
            "task",
            podman,
            cluster_quota,
            r"""
            apiVersion: v1
            kind: Pod
            spec:
              containers:
              - name: container
                image: busybox
                command:
                - "sh"
                - "-c"
                - |-
                    echo 'Hello world!'
                    echo "The message of the day is $(echo -n {{ repo0 }} | base64). That's great!"
                    echo 'Goodbye world!'
                    # lol {{ repo0 }}
                resources:                  
                  requests:                    
                    cpu: 10m
                    memory: 100Mi
            """,
            logs=repoLogs,
            done=repoDone,
        )
        task.link("repo0", repo0, is_input=True)

        pipeline = pydatatask.Pipeline([task], session)

        async with pipeline:
            await pydatatask.update(pipeline)
            live = task.links["live"].repo
            launched = [x async for x in live]
            assert launched
            assert await live.contains(launched[0])
            await pydatatask.delete_data(pipeline, "task", False, [launched[0]])
            while await live.contains(launched[0]):
                await asyncio.sleep(1)

            task.warned = False
            await task.launch(launched[0])
            await asyncio.sleep(5)
            assert await live.contains(launched[0])

            await pydatatask.run(pipeline, forever=False, launch_once=False, timeout=120)
            assert not await live.contains(launched[0])

        assert len(repo0.data) == 50
        for job, weh in repo0.data.items():
            assert "node" in repoDone.data[job]
            async with await repoLogs.open(job, "r") as fp:
                logs = await fp.read()
            assert (
                logs
                == f"""\
Hello world!
The message of the day is {base64.b64encode(weh.encode()).decode()}. That's great!
Goodbye world!
"""
            )

    async def asyncTearDown(self):
        if self.minikube_profile is not None:
            p = await asyncio.create_subprocess_exec(
                self.minikube_path,
                "delete",
                "--profile",
                self.minikube_profile,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT,
            )
            stdout, _ = await p.communicate()
            code = await p.wait()
            if code != 0:
                warnings.warn(f"minikube failed to delete {self.minikube_profile}. Logs below:\n\n{stdout.decode()}")


if __name__ == "__main__":
    unittest.main()



from typing import Optional
import asyncio
import os
import random
import shutil
import string
import unittest

import motor.motor_asyncio

import pydatatask


def rid(n=6):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(n))


class TestMongoDB(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.docker_name = None
        self.docker_path = shutil.which("docker")
        self.mongo_url = os.getenv("PYDATATASK_TEST_MONGODB_URL")
        self.test_id = rid()
        self.client: Optional[motor.motor_asyncio.AsyncIOMotorClient] = None
        self.database = "test-pydatatask-" + self.test_id

    async def asyncSetUp(self):
        if self.mongo_url is None:
            if self.docker_path is None:
                raise unittest.SkipTest("No mongodb endpoint configured and docker is not installed")
            port = random.randrange(0x4000, 0x8000)
            self.mongo_url = f"mongodb://root:root@localhost:{port}"
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "run",
                "--rm",
                "--name",
                self.database,
                "-d",
                "-p",
                f"{port}:27017",
                "-e",
                "MONGO_INITDB_ROOT_USERNAME=root",
                "-e",
                "MONGO_INITDB_ROOT_PASSWORD=root",
                "mongo:latest",
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()
            if await p.wait() != 0:
                raise unittest.SkipTest("No minio endpoint configured and docker failed to launch mongo:latest")
            self.docker_name = self.database
            await asyncio.sleep(1)
        self.client = motor.motor_asyncio.AsyncIOMotorClient(self.mongo_url)

    async def test_mongo(self):
        repo = pydatatask.MongoMetadataRepository(lambda: self.client[self.database], "test")
        assert repr(repo)
        await repo.dump("foo", {"weh": 1})
        assert len([x async for x in repo]) == 1
        assert (await repo.info("foo"))["weh"] == 1
        assert (await self.client[self.database].test.find_one({"_id": "foo"}))["weh"] == 1
        assert await repo.contains("foo")
        assert not await repo.contains("bar")
        all_things = await repo.info_all()
        assert len(all_things) == 1
        assert "foo" in all_things
        assert all_things["foo"]["weh"] == 1

        await repo.delete("foo")
        assert await repo.info("foo") == {}
        assert len([x async for x in repo]) == 0

    async def asyncTearDown(self):
        if self.client is not None:
            await self.client.drop_database(self.database)

        if self.docker_name is not None:
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "kill",
                self.docker_name,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
import asyncio
import base64
import random
import shutil
import string
import unittest

import aioshutil
import asyncssh

import pydatatask


def rid(n=6):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(n))


class TestLocalProcess(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.app = f"test-{rid()}"
        self.dir = Path(f"/tmp/pydatatask-{self.app}")
        self.n = 50

    async def test_local_process(self):
        session = pydatatask.Session()

        @session.resource
        async def localhost():
            yield pydatatask.LocalLinuxManager(self.app, local_path=self.dir)

        quota = pydatatask.ResourceManager(pydatatask.Resources.parse(1, 1))

        repo_input = pydatatask.FileRepository(self.dir / "input")
        await repo_input.validate()
        for i in range(self.n):
            async with await repo_input.open(str(i), "w") as fp:
                await fp.write(str(i))
        repo_stdout = pydatatask.InProcessBlobRepository()
        repo_done = pydatatask.InProcessMetadataRepository()
        repo_pids = pydatatask.InProcessMetadataRepository()

        template = """\
#!/bin/sh
echo weh | cat {{input}} -
echo bye >&2
        """

        task = pydatatask.ProcessTask(
            "task",
            localhost,
            quota,
            pydatatask.Resources.parse("100m", "100m"),
            repo_pids,
            template,
            {},
            repo_done,
            None,
            repo_stdout,
            pydatatask.STDOUT,
        )
        task.link("input", repo_input, is_input=True)

        pipeline = pydatatask.Pipeline([task], session)

        async with pipeline:
            await pydatatask.run(pipeline, False, False, 120)

        assert len(repo_pids.data) == 0
        assert len(repo_stdout.data) == self.n
        assert len(repo_done.data) == self.n

        for i in range(self.n):
            assert repo_done.data[str(i)]["return_code"] == 0
            assert repo_stdout.data[str(i)] == f"{i}weh\nbye\n".encode()

    async def asyncTearDown(self):
        await aioshutil.rmtree(self.dir, ignore_errors=True)


class TestSSHProcess(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.docker_name = None
        self.docker_path = shutil.which("docker")
        self.test_id = rid()
        self.port = None
        self.n = 50

    async def asyncSetUp(self):
        if self.docker_path is None:
            raise unittest.SkipTest("No mongodb endpoint configured and docker is not installed")
        self.port = random.randrange(0x4000, 0x8000)
        name = f"pydatatask-test-{self.test_id}"
        p = await asyncio.create_subprocess_exec(
            self.docker_path,
            "run",
            "--rm",
            "--name",
            name,
            "-d",
            "-p",
            f"{self.port}:2222",
            "-e",
            "USER_NAME=weh",
            "-e",
            "USER_PASSWORD=weh",
            "-e",
            "PASSWORD_ACCESS=true",
            "linuxserver/openssh-server:latest",
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await p.communicate()
        if await p.wait() != 0:
            raise unittest.SkipTest(
                "No minio endpoint configured and docker failed to launch linuxserver/openssh-server:latest"
            )
        self.docker_name = name
        await asyncio.sleep(2)

    async def test_ssh_process(self):
        session = pydatatask.Session()

        @session.resource
        async def ssh():
            async with asyncssh.connect(
                "localhost", port=self.port, username="weh", password="weh", known_hosts=None
            ) as s:
                yield s

        @session.resource
        async def procman():
            yield pydatatask.SSHLinuxManager(self.test_id, ssh)

        quota = pydatatask.ResourceManager(pydatatask.Resources.parse(1, 1))

        repo_stdin = pydatatask.InProcessBlobRepository({str(i): str(i).encode() for i in range(self.n)})
        repo_stdout = pydatatask.InProcessBlobRepository()
        repo_stderr = pydatatask.InProcessBlobRepository()
        repo_done = pydatatask.InProcessMetadataRepository()
        repo_pids = pydatatask.InProcessMetadataRepository()

        template = """\
#!/bin/sh
echo 'hello world!' >&2
echo "The message of day {{job}} is $(base64). That's pretty great!"
echo 'goodbye world!' >&2
        """

        task = pydatatask.ProcessTask(
            "task",
            procman,
            quota,
            pydatatask.Resources.parse("100m", "100m"),
            repo_pids,
            template,
            {},
            repo_done,
            repo_stdin,
            repo_stdout,
            repo_stderr,
        )

        pipeline = pydatatask.Pipeline([task], session)

        async with pipeline:
            await pydatatask.run(pipeline, False, False, 120)

        assert len(repo_stdin.data) == self.n
        assert len(repo_stdout.data) == self.n
        assert len(repo_stderr.data) == self.n
        assert len(repo_done.data) == self.n
        assert len(repo_pids.data) == 0

        for i in range(self.n):
            assert repo_stderr.data[str(i)] == f"hello world!\ngoodbye world!\n".encode()
            assert (
                repo_stdout.data[str(i)]
                == f"The message of day {i} is {base64.b64encode(repo_stdin.data[str(i)]).decode()}. That's pretty great!\n".encode()
            )
            assert repo_done.data[str(i)]["return_code"] == 0

    async def asyncTearDown(self):
        if self.docker_name is not None:
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "kill",
                self.docker_name,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest

import aiofiles.os
import aioshutil

import pydatatask


class TestFilesystem(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

        self.dir = None

    async def asyncSetUp(self):
        self.dir = tempfile.mkdtemp()

    async def test_filesystem(self):
        repo = pydatatask.FileRepository(self.dir, extension=".txt")
        assert repr(repo)
        await repo.validate()

        async with await repo.open("foo", "w") as fp:
            await fp.write("hello world")
        async with aiofiles.open(self.dir + "/foo.yaml", "w") as fp:
            await fp.write("# lol\n")
        async with aiofiles.open(self.dir + "/foo.txt", "rb") as fp:
            assert await fp.read() == b"hello world"
        async with await repo.open("foo", "rb") as fp:
            assert await fp.read() == b"hello world"
        assert [x async for x in repo] == ["foo"]
        assert await repo.info("bar") == self.dir + "/bar.txt"
        await repo.delete("foo")
        await repo.delete("bar")
        assert [x async for x in repo] == []

    async def test_case_insensitivity(self):
        repo = pydatatask.FileRepository(self.dir, extension=".txt", case_insensitive=True)
        await repo.validate()

        async with aiofiles.open(self.dir + "/foo.txt", "w") as fp:
            await fp.write("# lol\n")
        async with aiofiles.open(self.dir + "/bar.TXT", "w") as fp:
            await fp.write("# lol\n")
        assert {x async for x in repo} == {"foo", "bar"}

    async def test_directory(self):
        repo = pydatatask.DirectoryRepository(self.dir, discard_empty=False)
        repo_noempty = pydatatask.DirectoryRepository(self.dir, discard_empty=True)

        await repo.mkdir("foo")
        await repo.mkdir("foo")
        assert [x async for x in repo] == ["foo"]
        assert [x async for x in repo_noempty] == []
        assert await repo.contains("foo")
        assert not await repo.contains("bar")
        assert not await repo_noempty.contains("foo")
        assert not await repo_noempty.contains("bar")
        await aiofiles.os.mkdir(repo_noempty.fullpath("foo") / "weh")
        assert await repo_noempty.contains("foo")
        assert [x async for x in repo_noempty] == ["foo"]

        await repo.delete("foo")
        assert [x async for x in repo] == []

    async def test_yaml(self):
        repo = pydatatask.YamlMetadataFileRepository(self.dir)

        await repo.dump("foo", {"weh": 1})
        assert await repo.info("foo") == {"weh": 1}

    async def asyncTearDown(self):
        if self.dir is not None:
            await aioshutil.rmtree(self.dir)


if __name__ == "__main__":
    unittest.main()



import logging
import unittest

import pydatatask


class TestRepoBase(unittest.IsolatedAsyncioTestCase):
    async def test_derived(self):
        class DerivedRepository(pydatatask.Repository):
            async def unfiltered_iter(self):
                yield "foo"
                yield " "

            async def delete(self, job):
                pass

            async def info(self, job):
                return None

        repo = DerivedRepository()

        with self.assertLogs("pydatatask.repository") as cl:
            assert not await repo.contains("weh")
            assert await repo.contains("foo")
        assert len(cl.records) == 1
        for x in cl.records:
            assert "valid job" in x.message

    async def test_map(self):
        class DerivedRepository(pydatatask.Repository):
            async def unfiltered_iter(self):
                yield "foo"
                yield "bar"

            async def info(self, job):
                return "info"

            async def delete(self, key):
                logging.root.debug("deleting %s", key)

        repo = DerivedRepository()

        async def mapper(info):
            return info.upper()

        async def filter(job):
            return job == "foo"

        mapped = repo.map(mapper, filter, allow_deletes=True)

        assert [x async for x in mapped] == ["foo"]
        assert await mapped.contains("foo")
        assert not await mapped.contains("bar")
        assert not await mapped.contains("weh")

        assert await mapped.info("foo") == "INFO"
        assert await mapped.info("bar") == "INFO"
        assert await mapped.info("weh") == "INFO"

        with self.assertLogs(logging.root, "DEBUG") as cl:
            await mapped.delete("foo")
            await mapped.delete("bar")
            await mapped.delete("weh")
        assert [record.message for record in cl.records] == ["deleting foo", "deleting bar", "deleting weh"]

        assert await mapped.info_all() == {"foo": "INFO"}


if __name__ == "__main__":
    unittest.main()



from typing import Optional
import asyncio
import base64
import gc
import json
import random
import shutil
import socket
import string
import unittest

import docker_registry_client_async
import dxf

import pydatatask


def rid(n=6):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(n))


class TestDockerHub(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)
        self.client = None

    async def asyncSetUp(self):
        self.client = docker_registry_client_async.DockerRegistryClientAsync(
            client_session_kwargs={"connector_owner": True}, tcp_connector_kwargs={"family": socket.AF_INET}
        )

    async def test_docker_hub(self):
        repo = pydatatask.DockerRepository(lambda: self.client, "registry-1.docker.io", "library/registry")
        assert repr(repo)
        assert [x async for x in repo] != []

        with self.assertRaises(PermissionError):
            await repo.delete("latest")
        await self.client.add_credentials(credentials="asdf", endpoint="registry-1.docker.io")
        with self.assertRaises(dxf.exceptions.DXFUnauthorizedError):
            await repo.delete("latest")

    async def asyncTearDown(self):
        await self.client.close()
        gc.collect()


class TestDockerLocal(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)
        self.client: Optional[docker_registry_client_async.DockerRegistryClientAsync] = None
        self.docker_path = shutil.which("docker")
        self.docker_name = None
        self.test_id = rid()
        self.endpoint = None

    async def asyncSetUp(self):
        if self.docker_path is None:
            raise unittest.SkipTest("Docker is not installed")

        docker_registry_client_async.DockerRegistryClientAsync.DEFAULT_PROTOCOL = "http"
        self.client = docker_registry_client_async.DockerRegistryClientAsync(
            client_session_kwargs={"connector_owner": True}, tcp_connector_kwargs={"family": socket.AF_INET}, ssl=False
        )

        docker_name = f"pydatatask-test{self.test_id}"
        port = random.randrange(0x4000, 0x8000)
        p = await asyncio.create_subprocess_exec(
            self.docker_path,
            "run",
            "--rm",
            "--name",
            docker_name,
            "-e",
            "REGISTRY_STORAGE_DELETE_ENABLED=TRUE",
            "-e",
            f"REGISTRY_AUTH_HTPASSWD_REALM=http://localhost:{port}",
            "-e",
            "REGISTRY_AUTH_HTPASSWD_PATH=/htpasswd",
            "-d",
            "-p",
            f"{port}:5000",
            "library/registry:2",
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
        )
        await p.communicate()
        if await p.wait() != 0:
            raise unittest.SkipTest("Docker failed to launch registry:latest")

        await asyncio.sleep(1)

        p = await asyncio.create_subprocess_exec(
            self.docker_path,
            "logs",
            docker_name,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await p.communicate()
        if await p.wait() != 0:
            raise unittest.SkipTest("Docker failed to get logs for registry:latest")

        user = [line for line in stderr.split() if b"user=" in line][0].split(b"=")[1].decode()
        password = [line for line in stderr.split() if b"password=" in line][0].split(b"=")[1].decode().strip('"')

        self.docker_name = docker_name
        self.endpoint = f"localhost:{port}"
        await self.client.add_credentials(
            credentials=base64.b64encode(f"{user}:{password}".encode()).decode(), endpoint=self.endpoint
        )

    async def test_local_registry(self):
        repo = pydatatask.DockerRepository(lambda: self.client, self.endpoint, "foo/weh")
        info = await repo.info("bar")
        assert info["withdomain"] == f"{self.endpoint}/foo/weh:bar"
        assert info["withoutdomain"] == f"foo/weh:bar"
        image_name = docker_registry_client_async.ImageName("foo/weh", endpoint=self.endpoint, tag="bar")
        digest = docker_registry_client_async.FormattedSHA256.calculate(b"asdf")
        manifest = {
            "schemaVersion": 2,
            "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
            "config": {
                "mediaType": "text/plain",
                "size": 4,
                "digest": digest,
            },
            "layers": [],
        }

        x = await self.client.post_blob(image_name, data=b"asdf", digest=digest)
        await self.client.put_blob_upload(x.location, digest, data=b"asdf")
        await self.client.put_manifest(image_name, docker_registry_client_async.Manifest(json.dumps(manifest).encode()))

        assert [x async for x in repo] == ["bar"]
        await repo.delete("bar")
        assert [x async for x in repo] == []
        await repo.delete("bar")

    async def asyncTearDown(self):
        try:
            await self.client.close()
            gc.collect()
        finally:
            docker_registry_client_async.DockerRegistryClientAsync.DEFAULT_PROTOCOL = "https"

        if self.docker_name is not None:
            p = await asyncio.create_subprocess_exec(
                self.docker_path,
                "kill",
                self.docker_name,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.communicate()


if __name__ == "__main__":
    unittest.main()



import contextlib
import io
import unittest

import aiofiles.os

import pydatatask


class TestBasic(unittest.IsolatedAsyncioTestCase):
    def __init__(self, method):
        super().__init__(method)

    async def test_metadata(self):
        session = pydatatask.Session()
        repo0 = pydatatask.InProcessMetadataRepository()
        repo1 = pydatatask.InProcessBlobRepository()
        done = pydatatask.InProcessMetadataRepository()
        assert repr(repo0)
        assert repr(repo1)

        for i in range(100):
            repo0.data[str(i)] = i

        @pydatatask.InProcessSyncTask("task", done)
        async def task(
            job,
            repo0: pydatatask.MetadataRepository,
            repo1: pydatatask.BlobRepository,
        ):
            async with aiofiles.open("/dev/urandom", "rb") as fp:
                data = await fp.read(await repo0.info(job))
            async with await repo1.open(job, "w") as fp:
                await fp.write(data.hex()[: len(data)])

        task.link("repo0", repo0, is_input=True)
        task.link("repo1", repo1, is_output=True)

        pipeline = pydatatask.Pipeline([task], session)
        async with pipeline:
            await pydatatask.run(pipeline, forever=False, launch_once=False, timeout=None)

        assert len(repo0.data) == len(repo1.data)
        for job, i in repo0.data.items():
            assert len(repo1.data[job]) == i

        assert len([x async for x in repo0]) == 100
        await repo1.delete("0")
        assert len([x async for x in repo1]) == 99

        captured = io.StringIO()
        with contextlib.redirect_stdout(captured):
            await pydatatask.print_status(pipeline, False)
        assert captured.getvalue() == "task\ntask.repo0 100\ntask.done 100\ntask.repo1 99\n\n"

        captured = io.StringIO()
        with contextlib.redirect_stdout(captured):
            await pydatatask.print_trace(pipeline, False, ["0", "1", "foo"])
        assert captured.getvalue() == "task\ntask.repo0 0 1\ntask.done 0 1\ntask.repo1 1\n\n"


if __name__ == "__main__":
    unittest.main()



from concurrent.futures import ThreadPoolExecutor
import unittest

import pydatatask


class TestExecutor(unittest.IsolatedAsyncioTestCase):
    async def test_executor(self):
        session = pydatatask.Session()
        repo0 = pydatatask.InProcessMetadataRepository({"foo": "weh"})
        repo1 = pydatatask.InProcessMetadataRepository()
        repo2 = pydatatask.InProcessMetadataRepository()
        phase0_done = pydatatask.InProcessMetadataRepository()
        phase1_done = pydatatask.InProcessMetadataRepository()

        executor = ThreadPoolExecutor(1)

        @pydatatask.ExecutorTask("phase0", executor, phase0_done)
        async def phase0(
            job: str,
            repo_zero: pydatatask.InProcessMetadataRepository,
            repo_one: pydatatask.InProcessMetadataRepository,
        ):
            await repo_one.dump(job, await repo_zero.info(job) + "!")

        @pydatatask.ExecutorTask("phase1", executor, phase1_done)
        async def phase1(
            job: str, repo_one: pydatatask.InProcessMetadataRepository, repo_two: pydatatask.InProcessMetadataRepository
        ):
            await repo_two.dump(job, await repo_one.info(job) + "?")

        phase0.link("repo_zero", repo0, is_input=True)
        phase0.link("repo_one", repo1, is_output=True)

        phase1.plug(phase0)
        phase1.link("repo_two", repo2, is_output=True)

        pipeline = pydatatask.Pipeline([phase0, phase1], session)

        async with pipeline:
            await pydatatask.run(pipeline, forever=False, launch_once=False, timeout=120)
            await pydatatask.launch(pipeline, "phase0", "bar", sync=True, meta=True, force=True)

        assert repo2.data["foo"] == "weh!?"
        assert "exception" in phase0_done.data["bar"]


if __name__ == "__main__":
    unittest.main()
